@@ -1,4 +1,4 @@
-/* $Header: /cvsroot/osrs/libtiff/tools/tiff2ps.c,v 1.4 2001/09/08 17:28:08 warmerda Exp $ */
+/* $Header: /cvsroot/osrs/libtiff/tools/tiff2ps.c,v 1.19 2003/08/12 07:51:03 dron Exp $ */
 
 /*
  * Copyright (c) 1988-1997 Sam Leffler
@@ -28,6 +28,7 @@
 #include <stdlib.h>			/* for atof */
 #include <math.h>
 #include <time.h>
+#include <strings.h>
 
 #include "tiffio.h"
 
@@ -99,16 +100,21 @@
 #define	FALSE	0
 #endif
 
-int     ascii85 = FALSE;		/* use ASCII85 encoding */
+int	ascii85 = FALSE;		/* use ASCII85 encoding */
 int	interpolate = TRUE;		/* interpolate level2 image */
 int	level2 = FALSE;			/* generate PostScript level 2 */
+int	level3 = FALSE;			/* generate PostScript level 3 */
 int	printAll = FALSE;		/* print all images in file */
 int	generateEPSF = TRUE;		/* generate Encapsulated PostScript */
-int 	PSduplex = FALSE;		/* enable duplex printing */
+int	PSduplex = FALSE;		/* enable duplex printing */
 int	PStumble = FALSE;		/* enable top edge binding */
 int	PSavoiddeadzone = TRUE;		/* enable avoiding printer deadzone */
+float	maxPageHeight = 0;		/* maximum size to fit on page */
+float	splitOverlap = 0;		/* amount for split pages to overlag */
+int	rotate = FALSE;			/* rotate image by 180 degrees */
 char	*filename;			/* input filename */
 int	useImagemask = FALSE;		/* Use imagemask instead of image operator */
+uint16	res_unit = 0;			/* Resolution units: 1 - inches, 2 - cm*/
 
 /*
  * ASCII85 Encoding Support.
@@ -117,7 +123,7 @@ unsigned char ascii85buf[10];
 int	ascii85count;
 int	ascii85breaklen;
 
-int	TIFF2PS(FILE*, TIFF*, float, float);
+int	TIFF2PS(FILE*, TIFF*, float, float, double, double, int);
 void	PSpage(FILE*, TIFF*, uint32, uint32);
 void	PSColorContigPreamble(FILE*, uint32, uint32, int);
 void	PSColorSeparatePreamble(FILE*, uint32, uint32, int);
@@ -130,7 +136,7 @@ void	Ascii85Init(void);
 void	Ascii85Put(unsigned char code, FILE* fd);
 void	Ascii85Flush(FILE* fd);
 void    PSHead(FILE*, TIFF*, uint32, uint32, float, float, float, float);
-void 	PSTail(FILE*, int);
+void	PSTail(FILE*, int);
 
 #if	defined( EXP_ASCII85ENCODER)
 int Ascii85EncodeBlock( uint8 * ascii85_p, unsigned f_eod, const uint8 * raw_p, int raw_l );
@@ -142,6 +148,9 @@ int
 main(int argc, char* argv[])
 {
 	int dirnum = -1, c, np = 0;
+	int centered = 0;
+	float bottommargin = 0;
+	float leftmargin = 0;
 	float pageWidth = 0;
 	float pageHeight = 0;
 	uint32 diroff = 0;
@@ -149,12 +158,18 @@ main(int argc, char* argv[])
 	extern int optind;
 	FILE* output = stdout;
 
-	while ((c = getopt(argc, argv, "h:i:w:d:o:O:aemzps128DT")) != -1)
+	while ((c = getopt(argc, argv, "b:d:h:H:L:i:w:l:o:O:acelmrxyzps1238DT")) != -1)
 		switch (c) {
+		case 'b':
+			bottommargin = atof(optarg);
+			break;
+		case 'c':
+			centered = 1;
+			break;
 		case 'd':
 			dirnum = atoi(optarg);
 			break;
-	        case 'D':
+		case 'D':
 			PSduplex = TRUE;
 			break;
 		case 'i':
@@ -164,11 +179,19 @@ main(int argc, char* argv[])
 			PStumble = TRUE;
 			break;
 		case 'e':
+			PSavoiddeadzone = FALSE;
 			generateEPSF = TRUE;
 			break;
 		case 'h':
 			pageHeight = atof(optarg);
 			break;
+		case 'H':
+			maxPageHeight = atof(optarg);
+			if (pageHeight==0) pageHeight = maxPageHeight;
+			break;
+		case 'L':
+			splitOverlap = atof(optarg);
+			break;
 		case 'm':
 			useImagemask = TRUE;
 			break;
@@ -184,12 +207,18 @@ main(int argc, char* argv[])
 				exit(-2);
 			}
 			break;
+		case 'l':
+			leftmargin = atof(optarg);
+			break;
 		case 'a':
 			printAll = TRUE;
 			/* fall thru... */
 		case 'p':
 			generateEPSF = FALSE;
 			break;
+		case 'r':
+			rotate = TRUE;
+			break;
 		case 's':
 			printAll = FALSE;
 			break;
@@ -201,15 +230,26 @@ main(int argc, char* argv[])
 			break;
 		case '1':
 			level2 = FALSE;
+			level3 = FALSE;
 			ascii85 = FALSE;
 			break;
 		case '2':
 			level2 = TRUE;
 			ascii85 = TRUE;			/* default to yes */
 			break;
+		case '3':
+			level3 = TRUE;
+			ascii85 = TRUE;			/* default to yes */
+			break;
 		case '8':
 			ascii85 = FALSE;
 			break;
+		case 'x':
+			res_unit = RESUNIT_CENTIMETER;
+			break;
+		case 'y':
+			res_unit = RESUNIT_INCH;
+			break;
 		case '?':
 			usage(-1);
 		}
@@ -221,14 +261,15 @@ main(int argc, char* argv[])
 			else if (diroff != 0 &&
 			    !TIFFSetSubDirectory(tif, diroff))
 				return (-1);
-			np = TIFF2PS(output, tif, pageWidth, pageHeight);
+			np = TIFF2PS(output, tif, pageWidth, pageHeight,
+				leftmargin, bottommargin, centered);
 			TIFFClose(tif);
 		}
 	}
 	if (np)
-	        PSTail(output, np);
+		PSTail(output, np);
 	else
-	        usage(-1);
+		usage(-1);
 	if (output != stdout)
 		fclose(output);
 	return (0);
@@ -247,15 +288,14 @@ checkImage(TIFF* tif)
 {
 	switch (photometric) {
 	case PHOTOMETRIC_YCBCR:
-		if ((compression == COMPRESSION_JPEG 
-                     || compression == COMPRESSION_OJPEG)
-                     && planarconfiguration == PLANARCONFIG_CONTIG) {
+		if ((compression == COMPRESSION_JPEG || compression == COMPRESSION_OJPEG)
+			&& planarconfiguration == PLANARCONFIG_CONTIG) {
 			/* can rely on libjpeg to convert to RGB */
 			TIFFSetField(tif, TIFFTAG_JPEGCOLORMODE,
 				     JPEGCOLORMODE_RGB);
 			photometric = PHOTOMETRIC_RGB;
 		} else {
-			if (level2)
+			if (level2 || level3)
 				break;
 			TIFFError(filename, "Can not handle image with %s",
 			    "PhotometricInterpretation=YCbCr");
@@ -355,27 +395,36 @@ PhotoshopBanner(FILE* fd, uint32 w, uint32 h, int bs, int nc, char* startline)
 	fprintf(fd, "\"\n");
 }
 
+/*
+ *   pw : image width in pixels
+ *   ph : image height in pixels
+ * pprw : image width in PS units (72 dpi)
+ * pprh : image height in PS units (72 dpi)
+ */
 static void
 setupPageState(TIFF* tif, uint32* pw, uint32* ph, float* pprw, float* pprh)
 {
-	uint16 res_unit;
 	float xres, yres;
 
 	TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, pw);
 	TIFFGetField(tif, TIFFTAG_IMAGELENGTH, ph);
-	TIFFGetFieldDefaulted(tif, TIFFTAG_RESOLUTIONUNIT, &res_unit);
+	if (res_unit == 0)
+		TIFFGetFieldDefaulted(tif, TIFFTAG_RESOLUTIONUNIT, &res_unit);
 	/*
 	 * Calculate printable area.
 	 */
-	if (!TIFFGetField(tif, TIFFTAG_XRESOLUTION, &xres))
+	if (!TIFFGetField(tif, TIFFTAG_XRESOLUTION, &xres) || !xres)
 		xres = PS_UNIT_SIZE;
-	if (!TIFFGetField(tif, TIFFTAG_YRESOLUTION, &yres))
+	if (!TIFFGetField(tif, TIFFTAG_YRESOLUTION, &yres) || !yres)
 		yres = PS_UNIT_SIZE;
 	switch (res_unit) {
 	case RESUNIT_CENTIMETER:
 		xres *= 2.54, yres *= 2.54;
 		break;
+	case RESUNIT_INCH:
+		break;
 	case RESUNIT_NONE:
+	default:
 		xres *= PS_UNIT_SIZE, yres *= PS_UNIT_SIZE;
 		break;
 	}
@@ -396,21 +445,101 @@ isCCITTCompression(TIFF* tif)
 
 static	tsize_t tf_bytesperrow;
 static	tsize_t ps_bytesperrow;
-static 	tsize_t	tf_rowsperstrip;
+static	tsize_t	tf_rowsperstrip;
 static	tsize_t	tf_numberstrips;
 static	char *hex = "0123456789abcdef";
 
+/*
+ * imagewidth & imageheight are 1/72 inches
+ * pagewidth & pageheight are inches
+ */
+int
+PlaceImage(FILE *fp, float pagewidth, float pageheight,
+	   float imagewidth, float imageheight, int splitpage,
+	   double lm, double bm, int cnt)
+{
+	float xtran = 0;
+	float ytran = 0;
+	float xscale = 1;
+	float yscale = 1;
+	float left_offset = lm * PS_UNIT_SIZE;
+	float bottom_offset = bm * PS_UNIT_SIZE;
+	float subimageheight;
+	float splitheight;
+	float overlap;
+
+	pagewidth *= PS_UNIT_SIZE;
+	pageheight *= PS_UNIT_SIZE;
+
+	if (maxPageHeight==0)
+		splitheight = 0;
+	else
+		splitheight = maxPageHeight * PS_UNIT_SIZE;
+	overlap = splitOverlap * PS_UNIT_SIZE;
+
+	/*
+	 * WIDTH:
+	 *      if too wide, scrunch to fit
+	 *      else leave it alone
+	 */
+	if (imagewidth <= pagewidth) {
+		xscale = imagewidth;
+	} else {
+		xscale = pagewidth;
+	}
+
+	/* HEIGHT:
+	 *      if too long, scrunch to fit
+	 *      if too short, move to top of page
+	 */
+	if (imageheight <= pageheight) {
+		yscale = imageheight;
+		ytran = pageheight - imageheight;
+	} else if (imageheight > pageheight &&
+		(splitheight == 0 || imageheight <= splitheight)) {
+		yscale = pageheight;
+	} else /* imageheight > splitheight */ {
+		subimageheight = imageheight - (pageheight-overlap)*splitpage;
+		if (subimageheight <= pageheight) {
+			yscale = imageheight;
+			ytran = pageheight - subimageheight;
+			splitpage = 0;
+		} else if ( subimageheight > pageheight && subimageheight <= splitheight) {
+			yscale = imageheight * pageheight / subimageheight;
+			ytran = 0;
+			splitpage = 0;
+		} else /* sumimageheight > splitheight */ {
+			yscale = imageheight;
+			ytran = pageheight - subimageheight;
+			splitpage++;
+		}
+	}
+	
+	bottom_offset += ytran / (cnt?2:1);
+	if (cnt)
+	    left_offset += xtran / 2;
+	fprintf(fp, "%f %f translate\n", left_offset, bottom_offset);
+	fprintf(fp, "%f %f scale\n", xscale, yscale);
+	if (rotate)
+	    fputs ("1 1 translate 180 rotate\n", fp);
+
+	return splitpage;
+}
+
 
 /* returns the sequence number of the page processed */
 int
-TIFF2PS(FILE* fd, TIFF* tif, float pw, float ph)
+TIFF2PS(FILE* fd, TIFF* tif, float pw, float ph, double lm, double bm, int cnt)
 {
 	uint32 w, h;
 	float ox, oy, prw, prh;
-	float scale;
+	float scale = 1.0;
+	float left_offset = lm * PS_UNIT_SIZE;
+	float bottom_offset = bm * PS_UNIT_SIZE;
 	uint32 subfiletype;
 	uint16* sampleinfo;
 	static int npages = 0;
+	int split;
 
 	if (!TIFFGetField(tif, TIFFTAG_XPOSITION, &ox))
 		ox = 0;
@@ -419,12 +548,12 @@ TIFF2PS(FILE* fd, TIFF* tif, float pw, float ph)
 	setupPageState(tif, &w, &h, &prw, &prh);
 
 	do {
-	        tf_numberstrips = TIFFNumberOfStrips(tif);
+		tf_numberstrips = TIFFNumberOfStrips(tif);
 		TIFFGetFieldDefaulted(tif, TIFFTAG_ROWSPERSTRIP,
 		    &tf_rowsperstrip);
 		setupPageState(tif, &w, &h, &prw, &prh);
 		if (!npages)
-		        PSHead(fd, tif, w, h, prw, prh, ox, oy);
+			PSHead(fd, tif, w, h, prw, prh, ox, oy);
 		TIFFGetFieldDefaulted(tif, TIFFTAG_BITSPERSAMPLE,
 		    &bitspersample);
 		TIFFGetFieldDefaulted(tif, TIFFTAG_SAMPLESPERPIXEL,
@@ -456,19 +585,62 @@ TIFF2PS(FILE* fd, TIFF* tif, float pw, float ph)
 			tf_bytesperrow = TIFFScanlineSize(tif);
 			npages++;
 			fprintf(fd, "%%%%Page: %d %d\n", npages, npages);
+			if (!generateEPSF && ( level2 || level3 )) {
+				float psw,psh;
+				if (pw!=0 && ph!=0) {
+					psw=pw;
+					psh=ph;
+			    
+				}
+				else {
+					psw=rotate ? prh:prw;
+				    psh=rotate ? prw:prh;		    
+				}
+				fprintf(fd,
+	"1 dict begin /PageSize [ %f %f ] def currentdict end setpagedevice\n",
+				        psw, psh);
+				fputs(
+	"<<\n  /Policies <<\n    /PageSize 3\n  >>\n>> setpagedevice\n",
+				      fd);
+			}
 			fprintf(fd, "gsave\n");
 			fprintf(fd, "100 dict begin\n");
 			if (pw != 0 && ph != 0) {
-				/* NB: maintain image aspect ratio */
-				scale = (pw*PS_UNIT_SIZE/prw) < (ph*PS_UNIT_SIZE/prh) ?
-				    (pw*PS_UNIT_SIZE/prw) :
-				    (ph*PS_UNIT_SIZE/prh);
-				if (scale > 1.0)
-					scale = 1.0;
-				fprintf(fd, "0 %f translate\n", ph*PS_UNIT_SIZE-prh*scale);
-				fprintf(fd, "%f %f scale\n", prw*scale, prh*scale);
-			} else
+				if (maxPageHeight) { /* used -H option */
+					split = PlaceImage(fd,pw,ph,prw,prh,0,lm,bm,cnt);
+					while( split ) {
+					    PSpage(fd, tif, w, h);
+					    fprintf(fd, "end\n");
+					    fprintf(fd, "grestore\n");
+					    fprintf(fd, "showpage\n");
+					    npages++;
+					    fprintf(fd, "%%%%Page: %d %d\n", npages, npages);
+					    fprintf(fd, "gsave\n");
+					    fprintf(fd, "100 dict begin\n");
+					    split = PlaceImage(fd,pw,ph,prw,prh,split,lm,bm,cnt);
+					}
+				} else {
+					/* NB: maintain image aspect ratio */
+					scale = (pw*PS_UNIT_SIZE/prw) < (ph*PS_UNIT_SIZE/prh) ?
+					    (pw*PS_UNIT_SIZE/prw) :
+					    (ph*PS_UNIT_SIZE/prh);
+					if (scale > 1.0)
+						scale = 1.0;
+					bottom_offset +=
+						(ph * PS_UNIT_SIZE - prh * scale) / (cnt?2:1);
+					if (cnt)
+						left_offset += (pw * PS_UNIT_SIZE - prw * scale) / 2;
+					fprintf(fd, "%f %f translate\n",
+						left_offset, bottom_offset);
+					fprintf(fd, "%f %f scale\n", prw * scale, prh * scale);
+					if (rotate)
+						fputs ("1 1 translate 180 rotate\n", fd);
+				}
+			} else {
 				fprintf(fd, "%f %f scale\n", prw, prh);
+				if (rotate)
+					fputs ("1 1 translate 180 rotate\n", fd);
+			}
 			PSpage(fd, tif, w, h);
 			fprintf(fd, "end\n");
 			fprintf(fd, "grestore\n");
@@ -500,7 +672,7 @@ static char TumblePreamble[] = "\
 systemdict begin\n\
   /languagelevel where { pop languagelevel } { 1 } ifelse\n\
   2 ge { 1 dict dup /Tumble true put setpagedevice }\n\
-  { statusdict /settumble known { statusdict begin settumble true end } if\n\
+  { statusdict /settumble known { statusdict begin true settumble end } if\n\
   } ifelse\n\
 end\n\
 %%EndFeature\n\
@@ -520,7 +692,7 @@ void
 PSHead(FILE *fd, TIFF *tif, uint32 w, uint32 h, float pw, float ph,
 	float ox, float oy)
 {
-        time_t t;
+	time_t t;
 
 	(void) tif; (void) w; (void) h;
 	t = time(0);
@@ -533,16 +705,16 @@ PSHead(FILE *fd, TIFF *tif, uint32 w, uint32 h, float pw, float ph,
 	/* NB: should use PageBoundingBox */
 	fprintf(fd, "%%%%BoundingBox: 0 0 %ld %ld\n",
 	    (long) ceil(pw), (long) ceil(ph));
-	fprintf(fd, "%%%%LanguageLevel: %d\n", level2 ? 2 : 1);
+	fprintf(fd, "%%%%LanguageLevel: %d\n", (level3 ? 3 : (level2 ? 2 : 1)));
 	fprintf(fd, "%%%%Pages: (atend)\n");
 	fprintf(fd, "%%%%EndComments\n");
 	fprintf(fd, "%%%%BeginSetup\n");
 	if (PSduplex)
-	        fprintf(fd, "%s", DuplexPreamble);
+		fprintf(fd, "%s", DuplexPreamble);
 	if (PStumble)
-	        fprintf(fd, "%s", TumblePreamble);
-	if (PSavoiddeadzone && level2)
-	        fprintf(fd, "%s", AvoidDeadZonePreamble);
+		fprintf(fd, "%s", TumblePreamble);
+	if (PSavoiddeadzone && (level2 || level3))
+		fprintf(fd, "%s", AvoidDeadZonePreamble);
 	fprintf(fd, "%%%%EndSetup\n");
 }
 
@@ -550,7 +722,7 @@ void
 PSTail(FILE *fd, int npages)
 {
 	fprintf(fd, "%%%%Trailer\n");
-        fprintf(fd, "%%%%Pages: %d\n", npages);
+	fprintf(fd, "%%%%Pages: %d\n", npages);
 	fprintf(fd, "%%%%EOF\n");
 }
 
@@ -714,7 +886,21 @@ PS_Lvl2ImageDict(FILE* fd, TIFF* tif, uint32 w, uint32 h)
 	fputs(" <<\n", fd);
 	fputs("  /ImageType 1\n", fd);
 	fprintf(fd, "  /Width %lu\n", (unsigned long) tile_width);
-	fprintf(fd, "  /Height %lu\n", (unsigned long) tile_height);
+	/*
+	 * Workaround for some software that may crash when last strip
+	 * of image contains fewer number of scanlines than specified
+	 * by the `/Height' variable. So for stripped images with multiple
+	 * strips we will set `/Height' as `im_h', because one is 
+	 * recalculated for each strip - including the (smaller) final strip.
+	 * For tiled images and images with only one strip `/Height' will
+	 * contain number of scanlines in tile (or image height in case of
+	 * one-stripped image).
+	 */
+	if (TIFFIsTiled(tif) || tf_numberstrips == 1)
+		fprintf(fd, "  /Height %lu\n", (unsigned long) tile_height);
+	else
+		fprintf(fd, "  /Height im_h\n");
+	
 	if (planarconfiguration == PLANARCONFIG_SEPARATE && samplesperpixel > 1)
 		fputs("  /MultipleDataSources true\n", fd);
 	fprintf(fd, "  /ImageMatrix [ %lu 0 0 %ld %s %s ]\n",
@@ -841,7 +1027,7 @@ PS_Lvl2ImageDict(FILE* fd, TIFF* tif, uint32 w, uint32 h)
 			fputs("\t /BlackIs1 true\n", fd);
 		fprintf(fd, "\t>> /CCITTFaxDecode filter");
 		break;
-	case COMPRESSION_LZW:	/* 5: Lempel-Ziv  & Welch */
+	case COMPRESSION_LZW:	/* 5: Lempel-Ziv & Welch */
 		TIFFGetFieldDefaulted(tif, TIFFTAG_PREDICTOR, &predictor);
 		if (predictor == 2) {
 			fputs("\n\t<<\n", fd);
@@ -855,6 +1041,26 @@ PS_Lvl2ImageDict(FILE* fd, TIFF* tif, uint32 w, uint32 h)
 		}
 		fputs(" /LZWDecode filter", fd);
 		break;
+	case COMPRESSION_DEFLATE:	/* 5: ZIP */
+	case COMPRESSION_ADOBE_DEFLATE:
+		if ( level3 ) {
+			 TIFFGetFieldDefaulted(tif, TIFFTAG_PREDICTOR, &predictor);
+			 if (predictor > 1) {
+				fprintf(fd, "\t %% PostScript Level 3 only.");
+				fputs("\n\t<<\n", fd);
+				fprintf(fd, "\t /Predictor %u\n", predictor);
+				fprintf(fd, "\t /Columns %lu\n",
+					(unsigned long) tile_width);
+				fprintf(fd, "\t /Colors %u\n", samplesperpixel);
+					fprintf(fd, "\t /BitsPerComponent %u\n",
+					bitspersample);
+				fputs("\t>>", fd);
+			 }
+			 fputs(" /FlateDecode filter", fd);
+		} else {
+			use_rawdata = FALSE ;
+		}
+		break;
 	case COMPRESSION_PACKBITS:	/* 32773: Macintosh RLE */
 		fputs(" /RunLengthDecode filter", fd);
 		use_rawdata = TRUE;
@@ -874,7 +1080,6 @@ PS_Lvl2ImageDict(FILE* fd, TIFF* tif, uint32 w, uint32 h)
 	case COMPRESSION_NEXT:		/* 32766: NeXT 2-bit RLE */
 	case COMPRESSION_THUNDERSCAN:	/* 32809: ThunderScan RLE */
 	case COMPRESSION_PIXARFILM:	/* 32908: Pixar companded 10bit LZW */
-	case COMPRESSION_DEFLATE:	/* 32946: Deflate compression */
 	case COMPRESSION_JBIG:		/* 34661: ISO JBIG */
 		use_rawdata = FALSE;
 		break;
@@ -951,14 +1156,17 @@ PS_Lvl2page(FILE* fd, TIFF* tif, uint32 w, uint32 h)
 	tsize_t chunk_size, byte_count;
 
 #if defined( EXP_ASCII85ENCODER )
-	int			ascii85_l;		/* Length, in bytes, of ascii85_p[] data */
-	uint8		*	ascii85_p = 0;		/* Holds ASCII85 encoded data */
+	int			ascii85_l;	/* Length, in bytes, of ascii85_p[] data */
+	uint8		*	ascii85_p = 0;	/* Holds ASCII85 encoded data */
 #endif
 
 	PS_Lvl2colorspace(fd, tif);
 	use_rawdata = PS_Lvl2ImageDict(fd, tif, w, h);
 
+/* See http://bugzilla.remotesensing.org/show_bug.cgi?id=80 */
+#ifdef ENABLE_BROKEN_BEGINENDDATA
 	fputs("%%BeginData:\n", fd);
+#endif
 	fputs("exec\n", fd);
 
 	tiled_image = TIFFIsTiled(tif);
@@ -1068,7 +1276,7 @@ PS_Lvl2page(FILE* fd, TIFF* tif, uint32 w, uint32 h)
 		}
 
 		if ( !ascii85 ) {
-			if ( level2 )
+			if ( level2 || level3 )
 				putc( '>', fd );
 			putc('\n', fd);
 		}
@@ -1084,7 +1292,9 @@ PS_Lvl2page(FILE* fd, TIFF* tif, uint32 w, uint32 h)
 #endif
        
 	_TIFFfree(buf_data);
+#ifdef ENABLE_BROKEN_BEGINENDDATA
 	fputs("%%EndData\n", fd);
+#endif
 	return(TRUE);
 }
 
@@ -1096,7 +1306,7 @@ PSpage(FILE* fd, TIFF* tif, uint32 w, uint32 h)
 	if ( useImagemask && (bitspersample == 1) )
 		imageOp = "imagemask";
 
-	if (level2 && PS_Lvl2page(fd, tif, w, h))
+	if ((level2 || level3) && PS_Lvl2page(fd, tif, w, h))
 		return;
 	ps_bytesperrow = tf_bytesperrow;
 	switch (photometric) {
@@ -1124,7 +1334,7 @@ PSpage(FILE* fd, TIFF* tif, uint32 w, uint32 h)
 		fprintf(fd, "%s", RGBcolorimage);
 		PhotoshopBanner(fd, w, h, 1, 3, "false 3 colorimage");
 		fprintf(fd, "/scanLine %ld string def\n",
-		    (long) ps_bytesperrow);
+		    (long) ps_bytesperrow * 3L);
 		fprintf(fd, "%lu %lu 8\n",
 		    (unsigned long) w, (unsigned long) h);
 		fprintf(fd, "[%lu 0 0 -%lu 0 %lu]\n",
@@ -1360,8 +1570,8 @@ PSDataBW(FILE* fd, TIFF* tif, uint32 w, uint32 h)
 	tstrip_t s;
 
 #if defined( EXP_ASCII85ENCODER )
-	int			ascii85_l;		/* Length, in bytes, of ascii85_p[] data */
-	uint8		*	ascii85_p = 0;		/* Holds ASCII85 encoded data */
+	int	ascii85_l;		/* Length, in bytes, of ascii85_p[] data */
+	uint8	*ascii85_p = 0;		/* Holds ASCII85 encoded data */
 #endif
 
 	(void) w; (void) h;
@@ -1394,7 +1604,7 @@ PSDataBW(FILE* fd, TIFF* tif, uint32 w, uint32 h)
 #endif
 
 	if (ascii85)
-	        Ascii85Init();
+		Ascii85Init();
 
 	for (s = 0; s < TIFFNumberOfStrips(tif); s++) {
 		int cc = TIFFReadEncodedStrip(tif, s, tf_buf, stripsize);
@@ -1429,8 +1639,8 @@ PSDataBW(FILE* fd, TIFF* tif, uint32 w, uint32 h)
 
 	if ( !ascii85 )
 	{
-	    if ( level2 )
-	        fputs(">\n", fd);
+	    if ( level2 || level3)
+		fputs(">\n", fd);
 	}
 #if !defined( EXP_ASCII85ENCODER )
 	else
@@ -1519,7 +1729,7 @@ PSRawDataBW(FILE* fd, TIFF* tif, uint32 w, uint32 h)
 			fputs(">\n", fd);
 			breaklen = MAXLINE;
 		} else {
-		        Ascii85Init();
+			Ascii85Init();
 #if defined( EXP_ASCII85ENCODER )
 			ascii85_l = Ascii85EncodeBlock( ascii85_p, 1, tf_buf, cc );
 
@@ -1712,15 +1922,15 @@ int Ascii85EncodeBlock( uint8 * ascii85_p, unsigned f_eod, const uint8 * raw_p,
          * Output any straggler bytes:
          */
     
-        if ( raw_l )
+        if ( raw_l > 0 )
         {
             int             len;                /* Output this many bytes */
     
             len = raw_l + 1;
             val32 = *++raw_p << 24;             /* Prime the pump */
     
-            if ( --raw_l )  val32 += *(++raw_p) << 16;
-            if ( --raw_l )  val32 += *(++raw_p) <<  8;
+            if ( --raw_l > 0 )  val32 += *(++raw_p) << 16;
+            if ( --raw_l > 0 )  val32 += *(++raw_p) <<  8;
     
             val32 /= 85;
     
@@ -1759,23 +1969,32 @@ int Ascii85EncodeBlock( uint8 * ascii85_p, unsigned f_eod, const uint8 * raw_p,
 char* stuff[] = {
 "usage: tiff2ps [options] input.tif ...",
 "where options are:",
-" -1            generate PostScript Level I (default)",
-" -2            generate PostScript Level II",
-" -8            disable use of ASCII85 encoding with PostScript Level II",
+" -1            generate PostScript Level 1 (default)",
+" -2            generate PostScript Level 2",
+" -3            generate PostScript Level 3",
+" -8            disable use of ASCII85 encoding with PostScript Level 2/3",
+" -a            convert all directories in file (default is first)",
+" -b #          set the bottom margin to # inches",
+" -c            center image (-b and -l still add to this)",
 " -d #          convert directory number #",
 " -D            enable duplex printing (two pages per sheet of paper)",
-" -e            generate Encapsulated PostScript (EPS)",
+" -e            generate Encapsulated PostScript (EPS) (implies -z)",
 " -h #          assume printed page height is # inches (default 11)",
+" -w #          assume printed page width is # inches (default 8.5)",
+" -H #          split image if height is more than # inches",
+" -L #          overLap split images by # inches",
 " -i #          enable/disable (Nz/0) pixel interpolation (default: enable)",
+" -l #          set the left margin to # inches",
 " -m            use \"imagemask\" operator instead of \"image\"",
 " -o #          convert directory at file offset #",
 " -O file       write PostScript to file instead of standard output",
-" -a            convert all directories in file (default is first)",
 " -p            generate regular PostScript",
+" -r            rotate by 180 degrees",
 " -s            generate PostScript for a single image",
 " -T            print pages for top edge binding",
-" -w #          assume printed page width is # inches (default 8.5)",
-" -z            enable printing in the deadzone (only for PostScript Level II)",
+" -x            override resolution units as centimeters",
+" -y            override resolution units as inches",
+" -z            enable printing in the deadzone (only for PostScript Level 2/3)",
 NULL
 };
 
@@ -1786,6 +2005,7 @@ usage(int code)
 	int i;
 
 	setbuf(stderr, buf);
+        fprintf(stderr, "%s\n\n", TIFFGetVersion());
 	for (i = 0; stuff[i] != NULL; i++)
 		fprintf(stderr, "%s\n", stuff[i]);
 	exit(code);
